import 'dart:io';
import 'dart:math';

import 'package:dartz/dartz.dart';
import 'package:flutter_appauth/flutter_appauth.dart';
import 'package:freezed_annotation/freezed_annotation.dart';

import 'package:fhir/primitive_types/primitive_types.dart';
import 'package:fhir/dstu2.dart' as dstu2;
import 'package:fhir/stu3.dart' as stu3;
import 'package:fhir/r4.dart' as r4;
import 'package:fhir/r5.dart' as r5;

import 'enums/enums.dart';
import 'failures/auth_failure.dart';
import 'request/capabilities_request.dart';
import 'scopes/scopes.dart';

/// the star of our show, who you've all come to see, the Smart object who
/// will provide the client for interacting with the FHIR server
class FhirClient {
  FhirClient({
    this.version = FhirV.r4,
    @required this.baseUrl,
    @required this.clientId,
    @required this.redirectUri,
    this.launch,
    this.scopes,
    this.additionalParameters,
    this.authorize,
    this.token,
    this.secret,
  });

  /// specify which FHIR version you're working with, defaults to R4
  FhirV version;

  /// specify the baseUrl of the Capability Statement (or conformance
  /// statement for Dstu2). Note this may not be the same as the authentication
  /// server or the FHIR data server
  FhirUri baseUrl;

  /// the clientId of your app, must be pre-registered with the authorization
  /// server
  String clientId;

  /// the redurectUri of your app, must be pre-registered with the authorization
  /// server, need to follow the instructions from flutter_appauth
  /// https://pub.dev/packages/flutter_appauth
  /// about editing files for Android and iOS
  FhirUri redirectUri;

  /// if there are certain launch strings that need to be included
  String launch;

  /// the scopes that will be included with the request
  Scopes scopes;

  /// any additional parameters you'd like to pass as part of this request
  Map<String, String> additionalParameters = <String, String>{};

  /// the authorize Url from the Conformance/Capability Statement
  FhirUri authorize;

  /// the token Url from the Conformance/Capability Statement
  FhirUri token;

  String secret;

  String _accessToken;

  String _refreshToken;

  DateTime _accessTokenExpiration;

  Future<Map<String, String>> get authHeaders async {
    if (DateTime.now().isAfter(_accessTokenExpiration)) {
      final refresh = await _refresh;
      return refresh.fold(
        (l) => {},
        (r) => {HttpHeaders.authorizationHeader: 'Bearer $_accessToken'},
      );
    }
    return {HttpHeaders.authorizationHeader: 'Bearer $_accessToken'};
  }

  Future<Either<AuthFailure, Unit>> get _tokens async {
    /// this request simply includes all of the parameters we have to this
    /// point. The clientId, the redirect Url, the client secret, the
    /// authorize and token enpoints, a list of scopes, and if there are any
    /// other additional parameters are passed, they are included
    final request = secret == null
        ? AuthorizationTokenRequest(
            clientId,
            redirectUri.toString(),
            serviceConfiguration: AuthorizationServiceConfiguration(
              authorize.toString(),
              token.toString(),
            ),
            scopes: scopes.scopesList(),
          )
        : AuthorizationTokenRequest(
            clientId,
            redirectUri.toString(),
            clientSecret: secret,
            serviceConfiguration: AuthorizationServiceConfiguration(
              authorize.toString(),
              token.toString(),
            ),
            scopes: scopes.scopesList(),
          );
    request.additionalParameters = additionalParameters ?? <String, String>{};
    request.additionalParameters['nonce'] = nonce();
    request.additionalParameters['aud'] = baseUrl.toString();

    /// since we don't know what will happen
    try {
      print('trying authorization');

      /// call the authorizeAndExchangeCode method
      final authorization =
          await FlutterAppAuth().authorizeAndExchangeCode(request);

      print('received authorization');

      _accessToken = authorization.accessToken;
      _refreshToken = authorization.refreshToken;
      _accessTokenExpiration = authorization.accessTokenExpirationDateTime;
    } catch (e) {
      return left(AuthFailure.unknownFailure(failedValue: e));
    }

    return right(unit);
  }

  Future<Either<AuthFailure, Unit>> get _refresh async {
    final tokenRequest = secret == null
        ? TokenRequest(
            clientId,
            redirectUri.toString(),
            serviceConfiguration: AuthorizationServiceConfiguration(
              authorize.toString(),
              token.toString(),
            ),
            refreshToken: _refreshToken,
            grantType: 'refresh_token',
            scopes: scopes.scopesList(),
            issuer: clientId,
          )
        : TokenRequest(
            clientId,
            redirectUri.toString(),
            clientSecret: secret,
            serviceConfiguration: AuthorizationServiceConfiguration(
              authorize.toString(),
              token.toString(),
            ),
            refreshToken: _refreshToken,
            grantType: 'refresh_token',
            scopes: scopes.scopesList(),
            issuer: clientId,
          );
    tokenRequest.additionalParameters =
        additionalParameters ?? <String, String>{};
    tokenRequest.additionalParameters['nonce'] = nonce();

    try {
      print('refreshing authorization');
      final authorization = await FlutterAppAuth().token(tokenRequest);

      print('received authorization');

      _accessToken = authorization.accessToken;
      _refreshToken = authorization.refreshToken;
      _accessTokenExpiration = authorization.accessTokenExpirationDateTime;
    } catch (e) {
      return left(AuthFailure.unknownFailure(failedValue: e));
    }

    return right(unit);
  }

  Future<Either<AuthFailure, Unit>> get _getEndpoints async {
    /// because each version of FHIR will return a different type of Conformance
    /// or Capability Statement
    dynamic conformanceStatement;
    switch (version) {

      /// each case requests that particular version's capabilities or
      /// conformance statement, then searches for the token and authorize
      /// endpoint in it
      case FhirV.dstu2:
        {
          conformanceStatement =
              await CapabilitiesRequest.dstu2(base: Uri.parse('$baseUrl'))
                  .request();
          conformanceStatement.fold((l) => null, (r) {
            token = _getUri(r as dstu2.Conformance, 'token');
            authorize = _getUri(r as dstu2.Conformance, 'authorize');
          });
        }
        break;
      case FhirV.stu3:
        {
          conformanceStatement =
              await CapabilitiesRequest.stu3(base: Uri.parse('$baseUrl'))
                  .request();
          conformanceStatement.fold((l) => null, (r) {
            token = _getUri(r as stu3.CapabilityStatement, 'token');
            authorize = _getUri(r as stu3.CapabilityStatement, 'authorize');
          });
        }
        break;
      case FhirV.r4:
        {
          conformanceStatement =
              await CapabilitiesRequest.r4(base: Uri.parse('$baseUrl'))
                  .request();
          conformanceStatement.fold((l) => null, (r) {
            token = _getUri(r as r4.CapabilityStatement, 'token');
            authorize = _getUri(r as r4.CapabilityStatement, 'authorize');
          });
        }
        break;
      case FhirV.r5:
        {
          conformanceStatement =
              await CapabilitiesRequest.r5(base: Uri.parse('$baseUrl'))
                  .request();
          conformanceStatement.fold((l) => null, (r) {
            token = _getUri(r as r5.CapabilityStatement, 'token');
            authorize = _getUri(r as r5.CapabilityStatement, 'authorize');
          });
        }
        break;
    }

    /// if either authorize or token are still null, we return a failure
    if (authorize == null) {
      return left(AuthFailure.noAuthorizationEndpoint(baseUrl: baseUrl));
    }
    if (token == null) {
      return left(AuthFailure.noTokenEndpoint(baseUrl: baseUrl));
    }
    return right(unit);
  }

  /// the function when you're ready to request access, be sure to pass in the
  /// the client secret when you make a request if you're creating a confidential
  /// app
  Future<Either<AuthFailure, Unit>> access({
    String authUrl,
    String tokenUrl,
  }) async {
    if (authUrl == null || tokenUrl == null) {
      final endpoints = await _getEndpoints;
      if (endpoints.isLeft()) {
        return endpoints;
      }
    } else {
      authorize = FhirUri(authUrl);
      token = FhirUri(tokenUrl);
    }
    return _tokens;
  }

  /// convenience method for finding either the token or authorize endpoint
  FhirUri _getUri(dynamic capabilityStatement, String type) {
    if (capabilityStatement?.rest == null) {
      return null;
    } else if (capabilityStatement.rest[0]?.security?.extension_ == null) {
      return null;
    } else if (capabilityStatement.rest[0].security.extension_[0]?.extension_ ==
        null) {
      return null;
    } else {
      final statement = capabilityStatement
          .rest[0].security.extension_[0].extension_
          .firstWhere((ext) => ext.url.toString() == type, orElse: () => null);
      if (statement == null) {
        return statement;
      } else {
        return statement.valueUri;
      }
    }
  }
}

String nonce({int length}) {
  const _chars =
      'AaBbCcDdEeFfGgHhIiJjKkLlMmNnOoPpQqRrSsTtUuVvWwXxYyZz1234567890';
  final _rnd = Random();
  return String.fromCharCodes(Iterable.generate(
      length ?? 10, (_) => _chars.codeUnitAt(_rnd.nextInt(_chars.length))));
}
